/*
 * Lips_object_processing.js
 */

'use strict';

window.document.onreadystatechange = function () { // Called *TWO TIMES*: when "interactive" and later on... when "complete"
    if (window.document.readyState === "interactive") {
        window.console.log(window.document.readyState + ": DOM just loaded...");
        new Lips_object_processing();
    } else {
        if (window.document.readyState === "complete")
            window.console.log(window.document.readyState + ": All resources just loaded...");
    }
};

/*
 * dat.gui
 */
class Parameters {
    constructor() {
        this.Parameters = 'Test flag(s) below...';
        this.Color = 0;
        this.Texture = false;
    }
}

class Lips_object_processing {
    constructor() {
        window.document.body.innerWidth = 800;
        window.document.body.innerHeight = 600;

        this._lips_texture = new Promise(ready => {
            new THREE.TextureLoader().load('./img/Franck.jpg', texture => {
                texture.name = './img/Franck.jpg';
                texture.minFilter = THREE.LinearFilter;
                ready(texture);
            });
        });

        Promise.race([
            new Promise(ready => {
//                (new THREE.GLTFLoader()).load('./models/Lips.gltf', this._lips_as_GLTF_format.bind(this, ready));
            }),
            new Promise(ready => {
                (new THREE.OBJLoader()).load('./models/Lips.obj', this._lips_as_OBJ_format.bind(this, ready));
            })
        ]).then(function (value) {
            console.assert(value === "Lips as GLTF ready" || value === "Lips as OBJ ready");
            this._animate();
        }.bind(this));

        this._camera = new THREE.PerspectiveCamera(50, window.document.body.innerWidth / window.document.body.innerHeight, 0.1, 20000);
        this._camera.position.z = 15;
        this._renderer = new THREE.WebGLRenderer(/*{alpha: true}*/);
        this._renderer.setClearColor(new THREE.Color("rgb(255,0,0)"), 0.5);
        this._renderer.setPixelRatio((window.devicePixelRatio) ? window.devicePixelRatio : 1);
        this._renderer.setSize(window.innerWidth, window.innerHeight);
        this._scene = new THREE.Scene();
        this._scene.add(new THREE.AmbientLight(0xFFFFFF, 1)); // Required by Lips that comes with 'MeshPhongMaterial'

        // 'PlaneBufferGeometry(width, height, widthSegments, heightSegments)'
        let my_geometry = new THREE.PlaneBufferGeometry(400, 300, 1, 1);
// 'positions' has 12 elements if 'widthSegments === 1' and 'heightSegments === 1', i.e., 4 edges
        let positions = my_geometry.attributes.position.array;
        window.console.assert(positions instanceof Float32Array);
        window.console.log(JSON.stringify(positions)); // {"0":-200,"1":150,"2":0,"3":200,"4":150,"5":0,"6":-200,"7":-150,"8":0,"9":200,"10":-150,"11":0}

        this._tween = 0;
        window.document.body.appendChild(this._renderer.domElement);
    }
    _animate() {
        window.requestAnimationFrame(this._animate.bind(this));

        const lips = this._scene.getObjectByName('Lips geometry');
        if (lips)
            lips.rotateY(0.01);
        if (this._scene.getObjectByName('Scene'))
            this._scene.getObjectByName('Scene').rotateZ(0.01);
        if (this._scene.getObjectByName('LineSegments'))
            this._scene.getObjectByName('LineSegments').rotateZ(0.01);
        this._renderer.render(this._scene, this._camera);
    }

    _lips_as_GLTF_format(ready, lips) {
// window.alert(JSON.stringify(lips.scene));
        lips.scene.traverse(child => {
            window.console.assert(child.hasOwnProperty('isMesh') === false); // L'export GLTF a fait perdre le maillage initial !
//            window.alert(child.name + ' with type: ' + child.type); // Types: 'Scene' et 'LineSegments'... Bizarre ?
            child.name = child.type; // 'Scene' and 'LineSegments': new name because exporter set an arbitrary name...
            if (child.name === 'Scene')
                this._scene.add(child);
        });
        // Ready:
        ready("Lips as GLTF ready");
    }

    _lips_as_OBJ_format(ready, lips) {
        lips.traverse(child => { // 'traverse' explores 'lips' in a recursive way...
            if (child.type === "Mesh") {
// Centering must occur *BEFORE* morphing:
                child.geometry.center(); // Positioning...

                this._morphing(child);
                this._UV_mapping(child);
                this._scene.add(child);

                window.console.assert(child.material.type === "MeshPhongMaterial"); // Loaded 'Lips.obj' file comes with this...
                child.material.color.set(0xFFFFFF); // White
                child.material.morphTargets = true;
                child.material.side = THREE.DoubleSide;
                child.material.vertexColors = THREE.VertexColors; // Because of 'BufferGeometry'...
//                child.material.wireframe = true;
                // Scaling:
                child.scale.set(5, 5, 5); // Requires greater size...

                window.console.assert(child.geometry.type === "BufferGeometry");
                // Only 'child.geometry.attributes.position' is defined:
                // window.alert(JSON.stringify(child.geometry.attributes.position));
                // Coloring vertices:
                const pink = new THREE.Color("rgb(255,192,203)");
                const hotpink = new THREE.Color("rgb(255,105,180)");
                const deeppink = new THREE.Color("rgb(255,20,147)");
                const red = new THREE.Color(0xFF0000);
                let colors = new Float32Array(child.geometry.attributes.position.array.length);
                child.geometry.setAttribute('color', new THREE.BufferAttribute(colors, 3));
                // Outermost lips:
                for (let i = 0; i < 5760; i++) { // [0-5760[ included in 'child.geometry.attributes.color.count'
                    colors[ i * 3 ] = pink.r;
                    colors[ i * 3 + 1 ] = pink.g;
                    colors[ i * 3 + 2 ] = pink.b;
                }
                for (let i = 5760; i < 7488; i++) {
                    colors[ i * 3 ] = hotpink.r;
                    colors[ i * 3 + 1 ] = hotpink.g;
                    colors[ i * 3 + 2 ] = hotpink.b;
                }
                for (let i = 7488; i < 9216; i++) {
                    colors[ i * 3 ] = deeppink.r;
                    colors[ i * 3 + 1 ] = deeppink.g;
                    colors[ i * 3 + 2 ] = deeppink.b;
                }
                for (let i = 9216; i < 12672; i++) {
                    colors[ i * 3 ] = red.r;
                    colors[ i * 3 + 1 ] = red.g;
                    colors[ i * 3 + 2 ] = red.b;
                }

                /*
                 * dat.gui
                 */
                const dat_gui = new dat.GUI({width: 250});
                let parameters = new Parameters();
                let max = 5760; // Outermost lips
                dat_gui.add(parameters, 'Color', 0, max).step(1).onChange(this._color_outermost_lips.bind(this, child, max));
                dat_gui.add(parameters, 'Texture').onChange(function (lips, value) {
                    if (value)
                        this._lips_texture.then(texture => {
                            lips.material.map = texture; // UV mapping required...
                            lips.material.vertexColors = THREE.NoColors; // Avoid blending between vertices' color and material color...
                        });
                    else {
                        lips.material.map = null;
                        lips.material.vertexColors = THREE.VertexColors; // Because of 'BufferGeometry'...
                    }
                    lips.material.needsUpdate = true;
                }.bind(this, child));
                // Ready (Initial 'lips' object has 'Group' type composed of *only one* object with 'Mesh' type:
                ready("Lips as OBJ ready");
                return;
            }
        });
    }

    _color_outermost_lips(lips, max, value) {
        let colors = lips.geometry.getAttribute('color').array;
        const pink = new THREE.Color(0xFFC0CB);
        const white = new THREE.Color(0xFFFFFF);
        for (let i = 0; i < value; i++) {
            colors[ i * 3 ] = white.r;
            colors[ i * 3 + 1 ] = white.g;
            colors[ i * 3 + 2 ] = white.b;
        }
        for (let i = value; i < max; i++) {
            colors[ i * 3 ] = pink.r;
            colors[ i * 3 + 1 ] = pink.g;
            colors[ i * 3 + 2 ] = pink.b;
        }
        // 'BufferAttribute' object is re-sent to the GPU:
        lips.geometry.getAttribute('color').needsUpdate = true;
    }

    _morphing(lips) { // https://stackoverflow.com/questions/49000689/how-to-morphtarget-of-an-obj-file-buffergeometry
        // Clone the 'BufferAttribute':
        let position = new THREE.BufferAttribute(Float32Array.from(lips.geometry.attributes.position.array), lips.geometry.attributes.position.itemSize);
        lips.geometry.morphAttributes.position = [];
        lips.geometry.morphAttributes.position.push(position);

        for (let i = 0; i < 5760 / 2; i++) {
            position.setY(i, position.getY(i) + 1);
        }
        for (let i = 5760 / 2; i < 5760; i++) {
            position.setY(i, position.getY(i) - 1);
        }
        for (let i = 5760; i < 5760 + (7488 - 5760) / 2; i++) {
            position.setY(i, position.getY(i) + 0.5);
            position.setZ(i, position.getZ(i) + 1);
        }
        for (let i = 5760 + (7488 - 5760) / 2; i < 7488; i++) {
            position.setY(i, position.getY(i) - 0.5);
            position.setZ(i, position.getZ(i) + 1);
        }
        for (let i = 7488; i < 7488 + (9216 - 7488) / 2; i++) {
            position.setY(i, position.getY(i) + 0.25);
            position.setZ(i, position.getZ(i) + 0.5);
        }
        for (let i = 7488 + (9216 - 7488) / 2; i < 9216; i++) {
            position.setY(i, position.getY(i) - 0.25);
            position.setZ(i, position.getZ(i) + 0.5);
        }

        lips.updateMorphTargets(); // It *NO LONGER* works with 'THREE.Geometry'!
        lips.morphTargetInfluences[0] = 0;

        createjs.Tween.get(this, {onChange: function () {
                lips.morphTargetInfluences[0] = this._tween;
            }.bind(this), loop: true, paused: false})
                .to({_tween: 0.1}, 1000, createjs.Ease.linear)
                .to({_tween: 0}, 300, createjs.Ease.linear);
    }

    _UV_mapping(lips) { // https://stackoverflow.com/questions/19504337/three-js-buffergeometry-with-texture-coordinates
        window.console.assert(lips.geometry.attributes.position.count === 12672); // 12672 edges
// By definition:
        window.console.assert(lips.geometry.attributes.position.count === lips.geometry.attributes.position.array.length / 3);
        // Each vertex has one uv coordinate for texture mapping:
        let uvs = new Float32Array(lips.geometry.attributes.position.count * 2);
        lips.geometry.setAttribute('uv', new THREE.BufferAttribute(uvs, 2));

// Geomtry comes without faces' index:
// "If this attribute is not set, the renderer assumes that each three contiguous positions represent a single triangle. Default is null." 
        window.console.assert(lips.geometry.getIndex() === null);
        // https://stackoverflow.com/questions/36730365/how-can-i-add-faces-to-an-indexed-three-buffergeometry
// Faces require manuel def.:
//        let indices = new Uint8Array(?)
//        lips.geometry.setIndex(new THREE.BufferAttribute(indices, 1)); 

        lips.geometry.computeBoundingBox();
        const x_min = lips.geometry.boundingBox.min.x;
        const x_max = lips.geometry.boundingBox.max.x;
        const y_min = lips.geometry.boundingBox.min.y;
        const y_max = lips.geometry.boundingBox.max.y;

        for (let i = 0; i < lips.geometry.attributes.position.count; i++) {
            const x = lips.geometry.attributes.position.array[i * 3];
            const y = lips.geometry.attributes.position.array[i * 3 + 1];
            uvs[ i * 2 ] = (x - x_min) / (x_max - x_min);
            uvs[ i * 2 + 1 ] = (y - y_min) / (y_max - y_min);
        }
    }
}
